@@ -3,21 +3,19 @@ from pyreptasks import Switch
 ## Definition of the class DataFormatter.
 class DataFormatter:
     """
-    This class allows to define a specific format for different fields in a direct way. Afterwards, it is possible to use these formats to ensure that all impacted data respect them. Also the object checks that the data respects the specified format constraints.
+    This class allows to define a specific format for different fields in a direct way. Then, the object is able to convert some data string into the format specified by the fields, as well as to ensure that the data respects the fields constraints, otherwise it will throw an exception.
 
     An object of type DataFormatter does not need any argument during its declaration, as all the specifications will be done while adding fields and their formats to the object. All the data formatted, regardless of the choosen format, will delete innecesary spaces automatically (example: "   hello     world      " -> "hello world").
 
-    ATTRIBUTES:
-
-     - fields: A list containing the defined fields with the format options specified for each of them. It is created as an empty list when the object is created.
-
     METHODS:
 
      - add_field: This method allows the user to add a new field and the format it must respect to the object list 'fields'. Check the method description for detailed explanation.
 
      - delete_field: This method takes a field name as argument and remove the corresponding field from the object list 'fields', if it exists.
 
-     - show_fields: This method returns a list containing the name of each field in the object list 'fields'.
+     - show_fields: This method returns a list containing the name of each field defined in the formatter.
+
+     - field_format: This method takes the name of a field defined in the formatter as input and returns the format set for this field.
 
      - single_data_to_format: This method takes two arguments, the data to format and the field of which this data should be part. It returns the formatted data, following the formatting rules stablished by the field.
 
@@ -25,7 +23,7 @@ class DataFormatter:
     """
 
     def __init__(self):
-        self.fields = []
+        self.__fields = []
 
     def add_field(
         self,
@@ -36,6 +34,7 @@ class DataFormatter:
         float_field: bool = False,
         alphanumeric_field: bool = False,
         admitted: dict = {},
+        required: dict = {},
         length_range: list = [],
         value_range: list = [],
     ):
@@ -68,6 +67,16 @@ class DataFormatter:
 
         - admitted: <class 'dict'> Only available for alphabetic and alphanumeric fields. Its default value is an empty dict. It should be formed by the single characters that the formatter must skip while checking if the introduced data fits in the field's format specification as keys, and their corresponding values will be the maximum times they are admitted. If there is not limit in the amount of this characters, the value must be True. For example, if the field should be formed by alphabetic characters but admitting spaces, dots,commas and '@' just once. To ensure the data fits this description, it might be done using alphabetic_field = True and admitted = {" ":True,".":True, ",":True, "@":1}
 
+        - required: <class 'dict'> Not available for integer or float fields. Its default value is an empty dict. It may contain as keys:
+
+            - Single character strings: Meaning that this character must be present in the string.
+
+            - Two characters strings: Defining a range of characters, so the formatter checks that the string contains characters inside this range.
+
+            - Tuples of single characters: Defining a set of characters, so the formatter checks that the string contains characters inside this tuple.
+
+        The value of each key stands for the number of times the formatter has to check each key.
+
         - length_range: <class 'list'> This list contains lower and upper bounds on the data length. The formatter will ensure that the data respect these constraints. If an exact length is required, set both bounds equal to the desired length. The default value is an empty list, meaning that there is not any constraint to apply.
 
         - value_range: <class 'list'> Only available for integer or float fields. This list contains lower and upper bounds on the data value (only applied for numeric fields). The formatter will ensure that the data respect these constraints. If an exact value is required, set both bounds equal to the desired value. The default value is an empty list, meaning that there is not any constraint to apply.
@@ -144,15 +153,60 @@ class DataFormatter:
             raise ValueError(
                 "admitted values must be integers (if the admitted character is limited to a certain amount) or True (if there is not any restriction)"
             )
+        if required.__class__ != dict:
+            raise ValueError(
+                "required must be of type dict but {} was given.".format(
+                    required.__class__
+                )
+            )
+        if required != {} and (int_field or float_field):
+            raise ValueError(
+                "required must be an empty dict if the field is declared as integer or float."
+            )
+        if (
+            False
+            in [
+                required_key.__class__ in [str, tuple]
+                for required_key in required.keys()
+            ]
+            or False
+            in [
+                len(required_character) in [1, 2]
+                for required_character in required.keys()
+                if required_character.__class__ == str
+            ]
+            or False
+            in [
+                all(
+                    [element.__class__ == str and len(element) == 1 for element in tupl]
+                )
+                for tupl in required.keys()
+                if tupl.__class__ == tuple
+            ]
+            or False
+            in [
+                required_characters[0] <= required_characters[1]
+                for required_characters in required.keys()
+                if required_characters.__class__ == str
+                and len(required_characters) == 2
+            ]
+        ):
+            raise ValueError(
+                "required keys must be strings of at most 2 characters or tuples containing single characters. If you provide a two-range string, make sure that the second character is greater than the first one."
+            )
+        if False in [value.__class__ == int for value in required.values()]:
+            raise ValueError("required values must be integers.")
         if length_range.__class__ != list:
             raise ValueError(
                 "length_range must be of type list but {} given.".format(
                     length_range.__class__
                 )
             )
-        if len(length_range) not in [0, 2]:
+        if len(length_range) not in [0, 2] or False in [
+            item.__class__ == int for item in length_range
+        ]:
             raise ValueError(
-                "length_range must be kept empty if no range needed. Otherwise introduce the lower and upper bound."
+                "length_range must be kept empty if no range needed. Otherwise introduce the lower and upper bound as integers."
             )
         if value_range.__class__ != list:
             raise ValueError(
@@ -164,7 +218,9 @@ class DataFormatter:
             raise ValueError(
                 "value_range must be an empty list except if the field represents a number."
             )
-        if len(value_range) not in [0, 2]:
+        if len(value_range) not in [0, 2] or False in [
+            item.__class__ in [int, float] for item in value_range
+        ]:
             raise ValueError(
                 "value_range must not be modified if there is not need of bounds in a numeric field. Otherwise introduce the lower and upper bound."
             )
@@ -184,11 +240,13 @@ class DataFormatter:
             new_field["style"] = style_switch.exec(style)
         if len(admitted.keys()) > 0:
             new_field["admitted"] = admitted
+        if len(required.keys()) > 0:
+            new_field["required"] = required
         if len(length_range) == 2:
             new_field["length_range"] = length_range
         if len(value_range) == 2:
             new_field["value_range"] = value_range
-        self.fields.append(
+        self.__fields.append(
             new_field
         )  ## Once the new field is ready, add it to the formatter.
 
@@ -211,13 +269,42 @@ class DataFormatter:
                     field_name
                 )
             )
-        self.fields.pop(names.index(field_name))  ## Field deletion
+        self.__fields.pop(names.index(field_name))  ## Field deletion
 
     def show_fields(self):
         """
         This method returns a list with the names of the fields already defined in the formatter.
+
+        OUTPUT:
+
+        - <class 'list'> A list containing the name of the defined fields.
         """
-        return [field["name"] for field in self.fields]
+        return [field["name"] for field in self.__fields]
+
+    def field_format(self, field_name: str):
+        """
+        This method returns the format of the introduced field.
+
+        INPUTS:
+
+        - field_name: <class 'str'>. The name of the field whose format is to be returned.
+
+        OUTPUT:
+
+        <class 'dict'>. A dictionary containing the format of de specified field.
+        """
+        if field_name.__class__ != str:  ## Check that the field_name is well introduced
+            raise ValueError(
+                "field_name must be a string but {} given.".format(field_name.__class__)
+            )
+        names = self.show_fields()
+        if field_name not in names:  ## Check that the selected field actually exists.
+            raise NameError(
+                "the field {} is not defined in the formatter. Please use the method 'show_fields' to check the defined methods.".format(
+                    field_name
+                )
+            )
+        return self.__fields[names.index(field_name)]
 
     def single_data_to_format(self, data: str, field_name: str):
         """
@@ -248,7 +335,7 @@ class DataFormatter:
                     field_name
                 )
             )
-        field = self.fields[names.index(field_name)]
+        field = self.__fields[names.index(field_name)]
         ## Formatting data........
         if field["alphabetic_field"]:
             return self.__alpha_formatter(field, data)
@@ -279,22 +366,26 @@ class DataFormatter:
                     data.__class__
                 )
             )
-        if len(self.fields) != len(data):
+        if len(self.__fields) != len(data):
             raise ValueError(
                 "the introduced data is not valid. The reason is likely that the formatter is waiting for a different amount of data, please use the method 'show_fields' to check which fields the formatter is waiting for."
             )
         ## Formatting data........
         for index in range(len(data)):
-            if self.fields[index]["alphabetic_field"]:
-                data[index] = self.__alpha_formatter(self.fields[index], data[index])
-            elif self.fields[index]["int_field"]:
-                data[index] = self.__int_formatter(self.fields[index], data[index])
-            elif self.fields[index]["float_field"]:
-                data[index] = self.__float_formatter(self.fields[index], data[index])
-            elif self.fields[index]["alphanumeric_field"]:
-                data[index] = self.__alphanum_formatter(self.fields[index], data[index])
+            if self.__fields[index]["alphabetic_field"]:
+                data[index] = self.__alpha_formatter(self.__fields[index], data[index])
+            elif self.__fields[index]["int_field"]:
+                data[index] = self.__int_formatter(self.__fields[index], data[index])
+            elif self.__fields[index]["float_field"]:
+                data[index] = self.__float_formatter(self.__fields[index], data[index])
+            elif self.__fields[index]["alphanumeric_field"]:
+                data[index] = self.__alphanum_formatter(
+                    self.__fields[index], data[index]
+                )
             else:
-                data[index] = self.__default_formatter(self.fields[index], data[index])
+                data[index] = self.__default_formatter(
+                    self.__fields[index], data[index]
+                )
         return data
 
     ## This private method gives format to the introduced data when it should be formed just by alphabetic characters.
@@ -307,12 +398,26 @@ class DataFormatter:
                     alpha_string, alpha_string.__class__
                 )
             )
-        alpha_string = self.__string_cleaner(alpha_string)  ## Cleaning the string...
+        alpha_string = self.__string_cleaner(alpha_string)  ## Cleaning the string..
+        ## If we receive an empty string and it is admitted by the field format, nothing else to do.
+        try:
+            if field["length_range"][0] == 0 and len(alpha_string) == 0:
+                return alpha_string
+        except KeyError:
+            pass
         ## Apply the specific format and check the data fits it.
         try:
             pure_string = self.__remove_admitted(
                 field["admitted"], alpha_string
             )  ##The string once the admitted characters have been removed
+        except KeyError:
+            if not alpha_string.isalpha():
+                raise ValueError(
+                    "non-admitted character introduced in alphabetic field {}.".format(
+                        field["name"]
+                    )
+                )
+        else:
             if len(pure_string) == 0:
                 raise ValueError(
                     "the introduced data does not contain any alphabetic character in alphabetic field {}.".format(
@@ -325,13 +430,11 @@ class DataFormatter:
                         field["name"]
                     )
                 )
+        ## Check if the required characters have been included into the input data
+        try:
+            self.__check_required(field["required"], alpha_string)
         except KeyError:
-            if not alpha_string.isalpha():
-                raise ValueError(
-                    "non-admitted character introduced in alphabetic field {}.".format(
-                        field["name"]
-                    )
-                )
+            pass
         try:
             alpha_string = field["style"](alpha_string)
         except KeyError:
@@ -358,6 +461,12 @@ class DataFormatter:
         digit_string = self.__string_cleaner(
             digit_string
         )  ## Cleaning the string....have been removed
+        ## If we receive an empty string and it is admitted by the field format, nothing else to do.
+        try:
+            if field["length_range"][0] == 0 and len(digit_string) == 0:
+                return digit_string
+        except KeyError:
+            pass
         ## Apply the specific format and check the data fits it.
         try:
             int(digit_string)
@@ -394,6 +503,12 @@ class DataFormatter:
                 )
             )
         digit_string = self.__string_cleaner(digit_string)  ## Cleaning the string....
+        ## If we receive an empty string and it is admitted by the field format, nothing else to do.
+        try:
+            if field["length_range"][0] == 0 and len(digit_string) == 0:
+                return digit_string
+        except KeyError:
+            pass
         ## Apply the specific format and check the data fits it.
         if (
             digit_string.count(".") == 0
@@ -435,11 +550,25 @@ class DataFormatter:
         alphanum_string = self.__string_cleaner(
             alphanum_string
         )  ## Cleaning the string...
+        ## If we receive an empty string and it is admitted by the field format, nothing else to do.
+        try:
+            if field["length_range"][0] == 0 and len(alphanum_string) == 0:
+                return alphanum_string
+        except KeyError:
+            pass
         ## Apply the specific format and check the data fits it.
         try:
             pure_string = self.__remove_admitted(
                 field["admitted"], alphanum_string
             )  ##The string once the admitted characters have been removed
+        except KeyError:
+            if not alphanum_string.isalnum():
+                raise ValueError(
+                    "non-admitted character introduced in alphanumeric field {}.".format(
+                        field["name"]
+                    )
+                )
+        else:
             if len(pure_string) == 0:
                 raise ValueError(
                     "the introduced data does not contain any alphanumeric character in alphanumeric field {}.".format(
@@ -452,13 +581,11 @@ class DataFormatter:
                         field["name"]
                     )
                 )
+        ## Check if the required characters have been included into the input data
+        try:
+            self.__check_required(field["required"], alphanum_string)
         except KeyError:
-            if not alphanum_string.isalnum():
-                raise ValueError(
-                    "non-admitted character introduced in alphanumeric field {}.".format(
-                        field["name"]
-                    )
-                )
+            pass
         try:
             alphanum_string = field["style"](alphanum_string)
         except KeyError:
@@ -484,6 +611,11 @@ class DataFormatter:
             string = field["style"](string)
         except KeyError:
             pass
+        ## Check if the required characters have been included into the input data
+        try:
+            self.__check_required(field["required"], string)
+        except KeyError:
+            pass
         try:
             if len(string) not in range(
                 field["length_range"][0], field["length_range"][1] + 1
@@ -494,22 +626,61 @@ class DataFormatter:
         return string
 
     ## This private method removes admitted characters from the introduced data in order to check that the rest of the string fits the requirements. Also it ensures that the data does not exceed the amount of these admitted characters
-    def __remove_admitted(self, admitted: dict, string: str):
+    def __remove_admitted(self, admitted: dict, data: str):
         for admitted_character in admitted.keys():
             if (
                 admitted[admitted_character].__class__ == int
-                and string.count(admitted_character) > admitted[admitted_character]
+                and data.count(admitted_character) > admitted[admitted_character]
             ):
                 raise ValueError(
-                    "the introduced data {} must contain at most {} {}, but it contains {}.".format(
-                        string,
+                    "the introduced data {} may contain at most {} {}, but it contains {}.".format(
+                        data,
                         admitted[admitted_character],
                         admitted_character,
-                        string.count(admitted_character),
+                        data.count(admitted_character),
                     )
                 )
-            string = "".join(string.split(admitted_character))
-        return string
+            data = "".join(data.split(admitted_character))
+        return data
+
+    ## This private method checks that the introduced data contains all the required characters.
+    def __check_required(self, required: dict, data: str):
+        required_switch = Switch({str: self.__check_string, tuple: self.__check_tuple})
+        for required_key in required.keys():
+            required_switch.exec(required_key.__class__)(
+                required_key, required[required_key], data
+            )
+
+    ## Submethod of __check_required to use when the key is a string
+    def __check_string(self, required_key: str, required_value: int, data: str):
+        if len(required_key) == 1 and data.count(required_key) < required_value:
+            raise ValueError(
+                "the introduced data {} must contain the character {} at least {} times.".format(
+                    data, required_key, required_value
+                )
+            )
+        elif (
+            len(required_key) == 2
+            and [
+                True if required_key[0] <= character <= required_key[1] else False
+                for character in data
+            ].count(True)
+            < required_value
+        ):
+            raise ValueError(
+                "the introduced data {} must contain characters in the range {}-{} at least {} times.".format(
+                    data, required_key[0], required_key[1], required_value
+                )
+            )
+
+    ## Submethod of __check_required to use when the key is a tuple
+    def __check_tuple(self, required_key: tuple, required_value: int, data: str):
+        if sum([data.count(character) for character in required_key]) < required_value:
+            raise ValueError(
+                "the introduced data {} must contain characters from the set {} at least {} times.".format(
+                    data, required_key, required_value
+                )
+            )
 
     ## This private method "cleans" a string, deleting unnecessary spaces that must have been included by error.
     def __string_cleaner(self, string):
