@@ -20,11 +20,12 @@ def test_1():
     # - Last Name: alphabetic field admitting spaces, the first letter of each name should be capital.
     # - Short Description: alphabetic field of length at most 50 characters, admitting spaces, dots and commas, the first letter of the field should be capital
     # - Contact number: integer field of lenght exactly 9.
-    # - Contact mail: alphanumeric field, admitting @(once),_,.,-
+    # - Contact mail: alphanumeric field, admitting @(once),_,.,-. @ must appear in the data.
     # - Age: integer field, value must be in the range 18 -> 90
     # - Country: Alphabetic field, upper case for each letter.
     # - City: Alphabetic field, lower case for each letter.
     # - Monthly contribution: float field, not higher than 30.5
+    # - Password: alphabetic field admitting _,#,-,!,$, requiring a lower case, an upper case and requiring one of those special characters.
     # - Hash: No format
 
     my_formatter = DataFormatter()
@@ -46,6 +47,7 @@ def test_1():
         "Contact mail",
         alphanumeric_field=True,
         admitted={"@": 1, "_": True, ".": True, "-": True},
+        required={"@": 1},
     )
     my_formatter.add_field("Age", int_field=True, value_range=[18, 90])
     my_formatter.add_field(
@@ -57,6 +59,12 @@ def test_1():
     my_formatter.add_field(
         "Monthly contribution", float_field=True, value_range=[0.0, 30.5]
     )
+    my_formatter.add_field(
+        "password",
+        alphabetic_field=True,
+        admitted={"#": True, "$": True, "_": True, "-": True, "!": True},
+        required={"az": 1, "AZ": 1, ("_", "-", "!", "#", "$"): 1},
+    )
     my_formatter.add_field("Hash")
     # Tests start here
     assert my_formatter.single_data_to_format("  john   ", "Name") == "John"
@@ -87,6 +95,9 @@ def test_1():
     assert (
         my_formatter.single_data_to_format("12.99", "Monthly contribution") == "12.99"
     )
+    assert (
+        my_formatter.single_data_to_format("myPASSWORD$$", "password") == "myPASSWORD$$"
+    )
     assert (
         my_formatter.single_data_to_format("ksncpdcn344nsmdcpecped243m", "Hash")
         == "ksncpdcn344nsmdcpecped243m"
@@ -117,6 +128,7 @@ def test_2():
         "Contact mail",
         alphanumeric_field=True,
         admitted={"@": 1, "_": True, ".": True, "-": True},
+        required={"@": 1},
     )
     my_formatter.add_field("Age", int_field=True, value_range=[18, 90])
     my_formatter.add_field(
@@ -128,6 +140,12 @@ def test_2():
     my_formatter.add_field(
         "Monthly contribution", float_field=True, value_range=[0.0, 30.5]
     )
+    my_formatter.add_field(
+        "password",
+        alphabetic_field=True,
+        admitted={"#": True, "$": True, "_": True, "-": True, "!": True},
+        required={"az": 1, "AZ": 1, ("_", "-", "!", "#", "$"): 1},
+    )
     my_formatter.add_field("Hash")
     # Tests start here
     data = [
@@ -140,6 +158,7 @@ def test_2():
         "united kingdom",
         "London",
         "12.99",
+        "myPASSWORD$$",
         "ksncpdcn344nsmdcpecped243m",
     ]
     assert my_formatter.to_format(data) == [
@@ -152,6 +171,7 @@ def test_2():
         "UNITED KINGDOM",
         "london",
         "12.99",
+        "myPASSWORD$$",
         "ksncpdcn344nsmdcpecped243m",
     ]
 
@@ -312,7 +332,7 @@ def test_4():
         my_formatter.add_field("Monthly contribution", admitted={".": True})
     except ValueError:
         errors_encountered += 1
-    assert errors_encountered == 3  # Check that the exception is thrown in both cases.
+    assert errors_encountered == 3  # Check that the exception is thrown in all cases.
 
     # admitted containing a non-string key
     try:
@@ -329,6 +349,49 @@ def test_4():
         my_formatter.add_field("Name", alphabetic_field=True, admitted={".": "3"})
     except ValueError:
         pass
+    # required type error
+    try:
+        my_formatter.add_field("Name", required=3)
+    except ValueError:
+        pass
+    # required not allowed if int or float field
+    try:
+        my_formatter.add_field("Name", int_field=True, required={"#": 1})
+    except ValueError:
+        pass
+    try:
+        my_formatter.add_field("Name", float_field=True, required={"#": 1})
+    except ValueError:
+        pass
+    # required containing a non-string or tuple key
+    try:
+        my_formatter.add_field("Name", required={2: 1})
+    except ValueError:
+        pass
+    # required containing a tuple as key where not all the elements are single characters, or even strings
+    try:
+        my_formatter.add_field("Name", required={(2, "e"): 1})
+    except ValueError:
+        pass
+    try:
+        my_formatter.add_field("Name", required={("hi", "w"): 1})
+    except ValueError:
+        pass
+    # required containing a string key longer than 2 characters
+    try:
+        my_formatter.add_field("Name", required={"#wer": 1})
+    except ValueError:
+        pass
+    # required containing a two-character string key not respecting the string order
+    try:
+        my_formatter.add_field("Name", required={"ba": True})
+    except ValueError:
+        pass
+    # required containing a non-integer value
+    try:
+        my_formatter.add_field("Name", required={"a": "2"})
+    except ValueError:
+        pass
     # length_range type error
     try:
         my_formatter.add_field("Name", length_range=23)
@@ -684,7 +747,7 @@ def test_12():
     my_formatter.delete_field("Name")
 
 
-# Test 13: __remove_admitted private method exceptions (they would be thrown from a public method)
+# Test 13: __remove_admitted private method exceptions (they should be thrown from a public method)
 def test_13():
     my_formatter = DataFormatter()
     # An admitted character appears in the input, but it overtake the allowed amount
@@ -698,3 +761,72 @@ def test_13():
     except ValueError:
         pass
     my_formatter.delete_field("email")
+
+
+# Test 14: field_format method exceptions
+def test_14():
+    my_formatter = DataFormatter()
+    # field_name type error
+    try:
+        my_formatter.field_format(22)
+    except ValueError:
+        pass
+    # field_name does not match a defined field
+    try:
+        my_formatter.field_format("Name")
+    except NameError:
+        pass
+
+
+# Test 15: field_format method test
+def test_15():
+    my_formatter = DataFormatter()
+    my_formatter.add_field(
+        "Name",
+        alphabetic_field=True,
+        length_range=[10, 50],
+        admitted={" ": True},
+        style="title",
+    )
+    assert my_formatter.field_format("Name") == {
+        "name": "Name",
+        "alphabetic_field": True,
+        "int_field": False,
+        "float_field": False,
+        "alphanumeric_field": False,
+        "length_range": [10, 50],
+        "admitted": {" ": True},
+        "style": str.title,
+    }
+
+
+# Test 16: __check_required private method exception test (they should be thrown from a public method)
+def test_16():
+    my_formatter = DataFormatter()
+    my_formatter.add_field("Test", required={"#": 1})
+    try:
+        my_formatter.to_format(["single character not found"])
+    except ValueError:
+        pass
+    my_formatter.delete_field("Test")
+    my_formatter.add_field("Test", required={"az": 1})
+    try:
+        my_formatter.to_format(["THERE IS NOT ANY LOWER CASE HERE"])
+    except ValueError:
+        pass
+    my_formatter.delete_field("Test")
+    my_formatter.add_field("Test", required={("a", "e", "i", "o", "u"): 3})
+    try:
+        my_formatter.to_format(["N0 3 vowels"])
+    except ValueError:
+        pass
+
+
+# Test 17: Empty strings received and admitted
+def test_17():
+    my_formatter = DataFormatter()
+    my_formatter.add_field("Name", alphabetic_field=True, length_range=[0, 50])
+    my_formatter.add_field("Phone number", int_field=True, length_range=[0, 9])
+    my_formatter.add_field("Prize", float_field=True, length_range=[0, 50])
+    my_formatter.add_field("email", alphanumeric_field=True, length_range=[0, 30])
+    my_formatter.to_format(["", "", "", ""])
